/*
Copyright (c) 2014 Aaron Jensen

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

/*
Major changes for use with CUDA 5.5 - Aaron Jensen
Based on SmallPt GPU (OpenCL) - David Bucciarelli 2009

Change log:
1) Ported to CUDA 5.5

2) Combined functionality of: 
* rendering_kernel.cl
* geomfunc.h, and
* smallptGPU.c
.  into one CUDA (.cu) file

3) updated documentation for each function

4) important CUDA keywords on function headers:
__device__ - GPU function, not compiled to CPU. Called from GPU
__global__ - GPU function. Called from CPU

Development Environment:
* Visual Studio 2012 Ultimate
* CUDA 5.5
* GeForce GTX 765M
* Windows 8.1, 64-bit
*/

// used for pixel calculations, using unsigned char instead of floats
#define SCALAR 255

#include <stdio.h>
#include <iostream>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include <vector>

#include "camera.h"
#include "scene.h"
#include "displayfunc.h"
#include "geom.h"

// needed for cuda
#include <cuda_runtime.h>

// variables
static bool runTest = false;
static unsigned int *seeds;
Camera camera;
static int currentSample = 0;
Sphere *spheres;
unsigned int sphereCount;

unsigned int* d_iobuffer;
unsigned int* d_seeds;
static Sphere* d_spheres;
static Camera* d_camera;

__device__ bool drawHQ = true;

// colorWheelKernel()
// INPUT:   GPU allocated frame buffer, height and width of screen
// RETURN:  nothing
//   updates a pixel to cycle through the RGB color wheel. GPU function
//   called from CPU
__global__
   void colorWheelKernel(unsigned int* buffer, int width, int height) {
      // calculate position in array
      unsigned int offset = blockIdx.x * blockDim.x + threadIdx.x;

      // used to update 4 byts of an int
      unsigned char *  aByte;

      // point to beginning of desired pixel
      aByte =  (unsigned char *) &(buffer[offset]);  

      // update color
      switch (aByte[3]) {
      case 255: // Black -> Red
         aByte[0] = (aByte[0] + 1);
         aByte[3] = aByte[0] == 255 ? 254 : 255;
         break;
      case 254: // Red -> Yellow
         aByte[1] = (aByte[1] + 1);
         aByte[3] = aByte[1] == 255 ? 253 : 254;
         break;
      case 253: // Yellow -> Green
         aByte[0] = (aByte[0] - 1);
         aByte[3] = aByte[0] == 0 ? 252 : 253;
         break;
      case 252: // Green -> Cyan
         aByte[0] = aByte[0];
         aByte[1] = aByte[1];
         aByte[2] = (aByte[2] + 1);
         aByte[3] = aByte[2] == 255 ? 251 : 252;
         break;
      case 251: // Cyan -> Blue
         aByte[1] = (aByte[1] - 1);
         aByte[3] = aByte[1] == 0 ? 250 : 251;
         break;
      case 250: // Blue -> Magenta
         aByte[0] = (aByte[0] + 1);
         aByte[3] = aByte[0] == 255 ? 249 : 250;
         break;
      case 249: // Magenta -> Red
         aByte[2] = (aByte[2] - 1);
         aByte[3] = aByte[2] == 0 ? 254 : 249;
         break;
      default:
         if (aByte[0] != 0)  {
            aByte[0] = aByte[0] - 1;
         } else if (aByte[1] != 0) {
            aByte[1] = aByte[1] - 1;
         } else if (aByte[2] != 0) {
            aByte[2] = aByte[2] - 1;
         } else {
            aByte[3] = 255;
         }
      }
}

// getError()
// INPUT:   cuda error flag
// RETURN:  nothing
//   displays human readable CUDA error messages to console
void getError(cudaError_t err) {
   if(err != cudaSuccess) {
      std::cout << "Error " << cudaGetErrorString(err) << std::endl;
   }
}

// getRandom()
// INPUT:   two unsigned int seeds
// RETURN:  randomize float
//   calculates a random float on the GPU. GPU function only
__device__
   static float getRandom(unsigned int *seed0, unsigned int *seed1) {
      *seed0 = 36969 * ((*seed0) & 65535) + ((*seed0) >> 16);
      *seed1 = 18000 * ((*seed1) & 65535) + ((*seed1) >> 16);

      unsigned int ires = ((*seed0) << 16) + (*seed1);

      /* Convert to float */
      union {
         float f;
         unsigned int ui;
      } res;
      res.ui = (ires & 0x007fffff) | 0x40000000;

      return (res.f - 2.f) / 2.f;
}

// generateCameraRay()
// INPUT:   Camera object, two random seeds, height and width of screen
//          X and Y position of the pixel, references to Ray to generate
// RETURN:  nothing
//   generates a randomized ray from the camera through the pixel. GPU funciton only
__device__ static void generateCameraRay( Camera *camera, unsigned int *seed0, unsigned int *seed1,  int width,  int height, int x, int y, Ray *ray) {
   float invWidth = 1.f / width;
   float invHeight = 1.f / height;
   float r1 = getRandom(seed0, seed1) - .5f;
   float r2 = getRandom(seed0, seed1) - .5f;
   float kcx = (x + r1) * invWidth - .5f;
   float kcy = (y + r2) * invHeight - .5f;

   Vec rdir;
   vinit(rdir,
      camera->x.x * kcx + camera->y.x * kcy + camera->dir.x,
      camera->x.y * kcx + camera->y.y * kcy + camera->dir.y,
      camera->x.z * kcx + camera->y.z * kcy + camera->dir.z);

   Vec rorig;
   vsmul(rorig, 0.1f, rdir);
   vadd(rorig, rorig, camera->orig);

   vnorm(rdir);
   rinit(*ray, rorig, rdir);
}

// sphereIntersect()
// INPUT:   reference to a scene sphere and a ray to test
// RETURN:  distance from ray origin to sphere, 0 if no intersection
//   finds distance between ray origin and sphere. GPU function only
__device__
   float sphereIntersect(Sphere *s, Ray *r) {
      Vec op; /* Solve t^2*d.d + 2*t*(o-p).d + (o-p).(o-p)-R^2 = 0 */
      vsub(op, s->p, r->o);

      float b = vdot(op, r->d);
      float det = b * b - vdot(op, op) + s->rad * s->rad;
      if (det < 0.f)
         return 0.f;
      else
         det = sqrt(det);

      float t = b - det;
      if (t >  EPSILON)
         return t;
      else {
         t = b + det;

         if (t >  EPSILON)
            return t;
         else
            return 0.f;
      }
}

// uniformSampleSphere()
// INPUT:   random offsets, reference to a 3D point to set
// RETURN:  nothing
//   chooses a random point on a sphere, set to v. GPU function only
__device__
   void uniformSampleSphere(float u1, float u2, Vec *v) {
      float zz = 1.f - 2.f * u1;
      float r = sqrt(max(0.f, 1.f - zz * zz));
      float phi = 2.f * FLOAT_PI * u2;
      float xx = r * cos(phi);
      float yy = r * sin(phi);

      vinit(*v, xx, yy, zz);
}

// intersect()
// INPUT:   scene geometery, number of spheres, test ray, distance, reference to closest intersection
// RETURN:  if intersection exists
//   Finds closest intersection point. GPU function only
__device__
   int intersect(Sphere *spheres, unsigned int sphereCount, Ray *r, float *t, unsigned int *id) {
      float inf = (*t) = 1e20f;

      unsigned int i = sphereCount;
      for (; i--;) {
         float d = sphereIntersect(&spheres[i], r);
         if ((d != 0.f) && (d < *t)) {
            *t = d;
            *id = i;
         }
      }

      return (*t < inf);
}

// intersectP()
// INPUT:   scene geometery, sphere count, test ray, maximum depth
// RETURN:  nothing
//   Tests if an intersection exists. GPU function only
__device__
   int intersectP(Sphere *spheres, unsigned int sphereCount, Ray *r, float maxt) {
      unsigned int i = sphereCount;
      for (; i--;) {
         float d = sphereIntersect(&spheres[i], r);
         if ((d != 0.f) && (d < maxt))
            return 1;
      }

      return 0;
}

// sampleLights()
// INPUT:   scene geometery, number of spheres, random seeds, intersection point, surface normal, and resultant color
// RETURN:  nothing
//   Takes a lighting sample from each light. GPU function only
__device__
   void sampleLights(Sphere *spheres, unsigned int sphereCount, unsigned int *seed0, unsigned int *seed1, Vec *hitPoint, Vec *normal, Vec *result) {
      vclr(*result);

      /* For each light */
      unsigned int i;
      for (i = 0; i < sphereCount; i++) {
         Sphere *light = &spheres[i];
         if (!viszero(light->e)) {
            /* It is a light source */
            Ray shadowRay;
            shadowRay.o = *hitPoint;

            /* Choose a point over the light source */
            Vec unitSpherePoint;
            uniformSampleSphere(getRandom(seed0, seed1), getRandom(seed0, seed1), &unitSpherePoint);
            Vec spherePoint;
            vsmul(spherePoint, light->rad, unitSpherePoint);
            vadd(spherePoint, spherePoint, light->p);

            /* Build the shadow ray direction */
            vsub(shadowRay.d, spherePoint, *hitPoint);
            float len = sqrt(vdot(shadowRay.d, shadowRay.d));
            vsmul(shadowRay.d, 1.f / len, shadowRay.d);

            float wo = vdot(shadowRay.d, unitSpherePoint);
            if (wo > 0.f) {
               /* It is on the other half of the sphere */
               continue;
            } else
               wo = -wo;

            /* Check if the light is visible */
            float wi = vdot(shadowRay.d, *normal);
            if ((wi > 0.f) && (!intersectP(spheres, sphereCount, &shadowRay, len - EPSILON))) {
               Vec c; vassign(c, light->e);
               float s = (4.f * FLOAT_PI * light->rad * light->rad) * wi * wo / (len *len);
               vsmul(c, s, c);
               vadd(*result, *result, c);
            }
         }
      }
}

// radiancePathTracing()
// INPUT:   scene geometery, number of spheres, ray, random seeeds, and reference to result color
// RETURN:  nothing
//   performs color calculation for a pixel. GPU function only
__device__
   void radiancePathTracing (Sphere *spheres, unsigned int sphereCount, Ray *startRay, unsigned int *seed0, unsigned int *seed1, Vec *result) {
      // variables for path tracing
      Ray currentRay; { { ((currentRay).o).x = ((*startRay).o).x; ((currentRay).o).y = ((*startRay).o).y; ((currentRay).o).z = ((*startRay).o).z; }; { ((currentRay).d).x = ((*startRay).d).x; ((currentRay).d).y = ((*startRay).d).y; ((currentRay).d).z = ((*startRay).d).z; }; };
      Vec rad; { (rad).x = 0.f; (rad).y = 0.f; (rad).z = 0.f; };
      Vec throughput; { (throughput).x = 1.f; (throughput).y = 1.f; (throughput).z = 1.f; };

      unsigned int depth = 0;
      int specularBounce = 1;
      for (;; ++depth) {

         if (depth > 6) {
            *result = rad;
            return;
         }

         float t;
         unsigned int id = 0;
         if (!intersect(spheres, sphereCount, &currentRay, &t, &id)) {
            *result = rad;
            return;
         }

         Sphere *obj = &spheres[id];

         Vec hitPoint;
         { float k = (t); { (hitPoint).x = k * (currentRay.d).x; (hitPoint).y = k * (currentRay.d).y; (hitPoint).z = k * (currentRay.d).z; } };
         { (hitPoint).x = (currentRay.o).x + (hitPoint).x; (hitPoint).y = (currentRay.o).y + (hitPoint).y; (hitPoint).z = (currentRay.o).z + (hitPoint).z; };

         Vec normal;
         { (normal).x = (hitPoint).x - (obj->p).x; (normal).y = (hitPoint).y - (obj->p).y; (normal).z = (hitPoint).z - (obj->p).z; };
         { float l = 1.f / sqrt(((normal).x * (normal).x + (normal).y * (normal).y + (normal).z * (normal).z)); { float k = (l); { (normal).x = k * (normal).x; (normal).y = k * (normal).y; (normal).z = k * (normal).z; } }; };

         float dp = ((normal).x * (currentRay.d).x + (normal).y * (currentRay.d).y + (normal).z * (currentRay.d).z);

         Vec nl;

         float invSignDP = -1.f * sign(dp);
         { float k = (invSignDP); { (nl).x = k * (normal).x; (nl).y = k * (normal).y; (nl).z = k * (normal).z; } };


         Vec eCol; { (eCol).x = (obj->e).x; (eCol).y = (obj->e).y; (eCol).z = (obj->e).z; };
         if (!(((eCol).x == 0.f) && ((eCol).x == 0.f) && ((eCol).z == 0.f))) {
            if (specularBounce) {
               { float k = (fabs(dp)); { (eCol).x = k * (eCol).x; (eCol).y = k * (eCol).y; (eCol).z = k * (eCol).z; } };
               { (eCol).x = (throughput).x * (eCol).x; (eCol).y = (throughput).y * (eCol).y; (eCol).z = (throughput).z * (eCol).z; };
               { (rad).x = (rad).x + (eCol).x; (rad).y = (rad).y + (eCol).y; (rad).z = (rad).z + (eCol).z; };
            }

            *result = rad;
            return;
         }

         // diffuse material
         if (obj->refl == DIFF) {
            specularBounce = 0;
            { (throughput).x = (throughput).x * (obj->c).x; (throughput).y = (throughput).y * (obj->c).y; (throughput).z = (throughput).z * (obj->c).z; };

            Vec Ld;
            sampleLights(spheres, sphereCount, seed0, seed1, &hitPoint, &nl, &Ld);
            { (Ld).x = (throughput).x * (Ld).x; (Ld).y = (throughput).y * (Ld).y; (Ld).z = (throughput).z * (Ld).z; };
            { (rad).x = (rad).x + (Ld).x; (rad).y = (rad).y + (Ld).y; (rad).z = (rad).z + (Ld).z; };

            float r1 = 2.f * 3.14159265358979323846f * getRandom(seed0, seed1);
            float r2 = getRandom(seed0, seed1);
            float r2s = sqrt(r2);

            Vec w; { (w).x = (nl).x; (w).y = (nl).y; (w).z = (nl).z; };

            Vec u, a;
            if (fabs(w.x) > .1f) {
               { (a).x = 0.f; (a).y = 1.f; (a).z = 0.f; };
            } else {
               { (a).x = 1.f; (a).y = 0.f; (a).z = 0.f; };
            }
            { (u).x = (a).y * (w).z - (a).z * (w).y; (u).y = (a).z * (w).x - (a).x * (w).z; (u).z = (a).x * (w).y - (a).y * (w).x; };
            { float l = 1.f / sqrt(((u).x * (u).x + (u).y * (u).y + (u).z * (u).z)); { float k = (l); { (u).x = k * (u).x; (u).y = k * (u).y; (u).z = k * (u).z; } }; };

            Vec v;
            { (v).x = (w).y * (u).z - (w).z * (u).y; (v).y = (w).z * (u).x - (w).x * (u).z; (v).z = (w).x * (u).y - (w).y * (u).x; };

            Vec newDir;
            { float k = (cos(r1) * r2s); { (u).x = k * (u).x; (u).y = k * (u).y; (u).z = k * (u).z; } };
            { float k = (sin(r1) * r2s); { (v).x = k * (v).x; (v).y = k * (v).y; (v).z = k * (v).z; } };
            { (newDir).x = (u).x + (v).x; (newDir).y = (u).y + (v).y; (newDir).z = (u).z + (v).z; };
            { float k = (sqrt(1 - r2)); { (w).x = k * (w).x; (w).y = k * (w).y; (w).z = k * (w).z; } };
            { (newDir).x = (newDir).x + (w).x; (newDir).y = (newDir).y + (w).y; (newDir).z = (newDir).z + (w).z; };

            currentRay.o = hitPoint;
            currentRay.d = newDir;
            continue;
         }
         
         // shiny material
         else if (obj->refl == SPEC) {
            specularBounce = 1;

            Vec newDir;
            { float k = (2.f * ((normal).x * (currentRay.d).x + (normal).y * (currentRay.d).y + (normal).z * (currentRay.d).z)); { (newDir).x = k * (normal).x; (newDir).y = k * (normal).y; (newDir).z = k * (normal).z; } };
            { (newDir).x = (currentRay.d).x - (newDir).x; (newDir).y = (currentRay.d).y - (newDir).y; (newDir).z = (currentRay.d).z - (newDir).z; };

            { (throughput).x = (throughput).x * (obj->c).x; (throughput).y = (throughput).y * (obj->c).y; (throughput).z = (throughput).z * (obj->c).z; };

            { { ((currentRay).o).x = (hitPoint).x; ((currentRay).o).y = (hitPoint).y; ((currentRay).o).z = (hitPoint).z; }; { ((currentRay).d).x = (newDir).x; ((currentRay).d).y = (newDir).y; ((currentRay).d).z = (newDir).z; }; };
            continue;
         }
         
         // refractive material
         else {
            specularBounce = 1;

            Vec newDir;
            { float k = (2.f * ((normal).x * (currentRay.d).x + (normal).y * (currentRay.d).y + (normal).z * (currentRay.d).z)); { (newDir).x = k * (normal).x; (newDir).y = k * (normal).y; (newDir).z = k * (normal).z; } };
            { (newDir).x = (currentRay.d).x - (newDir).x; (newDir).y = (currentRay.d).y - (newDir).y; (newDir).z = (currentRay.d).z - (newDir).z; };

            Ray reflRay; { { ((reflRay).o).x = (hitPoint).x; ((reflRay).o).y = (hitPoint).y; ((reflRay).o).z = (hitPoint).z; }; { ((reflRay).d).x = (newDir).x; ((reflRay).d).y = (newDir).y; ((reflRay).d).z = (newDir).z; }; };
            int into = (((normal).x * (nl).x + (normal).y * (nl).y + (normal).z * (nl).z) > 0);

            float nc = 1.f;
            float nt = 1.5f;
            float nnt = into ? nc / nt : nt / nc;
            float ddn = ((currentRay.d).x * (nl).x + (currentRay.d).y * (nl).y + (currentRay.d).z * (nl).z);
            float cos2t = 1.f - nnt * nnt * (1.f - ddn * ddn);

            if (cos2t < 0.f) {
               { (throughput).x = (throughput).x * (obj->c).x; (throughput).y = (throughput).y * (obj->c).y; (throughput).z = (throughput).z * (obj->c).z; };

               { { ((currentRay).o).x = ((reflRay).o).x; ((currentRay).o).y = ((reflRay).o).y; ((currentRay).o).z = ((reflRay).o).z; }; { ((currentRay).d).x = ((reflRay).d).x; ((currentRay).d).y = ((reflRay).d).y; ((currentRay).d).z = ((reflRay).d).z; }; };
               continue;
            }

            float kk = (into ? 1 : -1) * (ddn * nnt + sqrt(cos2t));
            Vec nkk;
            { float k = (kk); { (nkk).x = k * (normal).x; (nkk).y = k * (normal).y; (nkk).z = k * (normal).z; } };
            Vec transDir;
            { float k = (nnt); { (transDir).x = k * (currentRay.d).x; (transDir).y = k * (currentRay.d).y; (transDir).z = k * (currentRay.d).z; } };
            { (transDir).x = (transDir).x - (nkk).x; (transDir).y = (transDir).y - (nkk).y; (transDir).z = (transDir).z - (nkk).z; };
            { float l = 1.f / sqrt(((transDir).x * (transDir).x + (transDir).y * (transDir).y + (transDir).z * (transDir).z)); { float k = (l); { (transDir).x = k * (transDir).x; (transDir).y = k * (transDir).y; (transDir).z = k * (transDir).z; } }; };

            float a = nt - nc;
            float b = nt + nc;
            float R0 = a * a / (b * b);
            float c = 1 - (into ? -ddn : ((transDir).x * (normal).x + (transDir).y * (normal).y + (transDir).z * (normal).z));

            float Re = R0 + (1 - R0) * c * c * c * c*c;
            float Tr = 1.f - Re;
            float P = .25f + .5f * Re;
            float RP = Re / P;
            float TP = Tr / (1.f - P);

            if (getRandom(seed0, seed1) < P) {
               { float k = (RP); { (throughput).x = k * (throughput).x; (throughput).y = k * (throughput).y; (throughput).z = k * (throughput).z; } };
               { (throughput).x = (throughput).x * (obj->c).x; (throughput).y = (throughput).y * (obj->c).y; (throughput).z = (throughput).z * (obj->c).z; };

               { { ((currentRay).o).x = ((reflRay).o).x; ((currentRay).o).y = ((reflRay).o).y; ((currentRay).o).z = ((reflRay).o).z; }; { ((currentRay).d).x = ((reflRay).d).x; ((currentRay).d).y = ((reflRay).d).y; ((currentRay).d).z = ((reflRay).d).z; }; };
               continue;
            } else {
               { float k = (TP); { (throughput).x = k * (throughput).x; (throughput).y = k * (throughput).y; (throughput).z = k * (throughput).z; } };
               { (throughput).x = (throughput).x * (obj->c).x; (throughput).y = (throughput).y * (obj->c).y; (throughput).z = (throughput).z * (obj->c).z; };

               { { ((currentRay).o).x = (hitPoint).x; ((currentRay).o).y = (hitPoint).y; ((currentRay).o).z = (hitPoint).z; }; { ((currentRay).d).x = (transDir).x; ((currentRay).d).y = (transDir).y; ((currentRay).d).z = (transDir).z; }; };
               continue;
            }
         }
      }
}

// radianceKernel()
// INPUT:   io pixel buffer, random seeds, camera, scene geometry, width and height
//          of frame buffer, number of spheres and the current sample count
// RETURN:  nothing
//   Sets color of a pixel. GPU function called from CPU
__global__
   void radianceKernel(unsigned int* pixBuffer, unsigned int* seedBuffer, Camera* camera, Sphere* spheres,  int width,  int height, int numSpheres, int curSample) {
      // calculate offset in 1D array and x,y position on the screen of this thread
      unsigned int offset = blockIdx.x * blockDim.x + threadIdx.x;
      unsigned int offset2 = 2 * offset;
      int x = offset % width;
      int y = offset / width;

      // access individual bytes of pixBuffer
      unsigned char * aByte;
      aByte =  (unsigned char *) &(pixBuffer[offset]);

      /* Check if we have to do something */
      if (y >= height)
         return;

      /* LordCRC: move seed to local store */
      unsigned int seed0 = seedBuffer[offset2];
      unsigned int seed1 = seedBuffer[offset2 + 1];

      // shoot a ray
      Ray ray1;
      Ray ray2;
      generateCameraRay(camera, &seed0, &seed1, width, height, x, y, &ray1);
      if (drawHQ) generateCameraRay(camera, &seed1, &seed0, width, height, x, y, &ray2);

      // find color of pixel
      Vec r1;
      Vec r2;
      radiancePathTracing(spheres, numSpheres, &ray1, &seed0, &seed1, &r1);
      if (drawHQ) {
         radiancePathTracing(spheres, numSpheres, &ray2, &seed1, &seed0, &r2);

         r1.x *= 0.5f + 0.5f * r2.x;
         r1.y *= 0.5f + 0.5f * r2.y;
         r1.z *= 0.5f + 0.5f * r2.z;
      }

      // copy calculated color to pixBuffer by reference
      if (curSample == 0) {
         aByte[0] = min(SCALAR * r1.x, SCALAR);
         aByte[1] = min(SCALAR * r1.y, SCALAR);
         aByte[2] = min(SCALAR * r1.z, SCALAR);
      } else {
         float k1 = (float) curSample;
         float k2 = 1.f / (curSample + 1.f);
         aByte[0] = min((aByte[0] * k1  + SCALAR * r1.x) * k2 + 0.5f, SCALAR);
         aByte[1] = min((aByte[1] * k1  + SCALAR * r1.y) * k2 + 0.5f, SCALAR);
         aByte[2] = min((aByte[2] * k1  + SCALAR * r1.z) * k2 + 0.5f, SCALAR);
      }
      aByte[3] = 255;

      // change used seeds
      seedBuffer[offset2] = seed0;
      seedBuffer[offset2 + 1] = seed1;
}

// freeBuffers()
// INPUT:   nothing
// RETURN:  nothing
//   releases used GPU memory. CPU function
static void freeBuffers() {
   // free device memory
   getError(cudaFree(d_iobuffer));
   getError(cudaFree(d_seeds));
}

// allocateBuffers()
// INPUT:   nothing
// RETURN:  nothing
//   Allocates GPU memory. CPU function
static void allocateBuffers() {
   // number of pixels (needed?)
   const int pixelCount = width * height;

   // iterator
   int i;

   // resize and fill buffer
   frameBuffer.resize(pixelCount);
   for (i=0; i<frameBuffer.size(); i++) {
      frameBuffer[i] = runTest ? i : 0;
   }

   // allocate HOST memory 
   seeds = (unsigned int *)malloc(sizeof(unsigned int) * pixelCount * 2);
   for (i = 0; i < pixelCount * 2; i++) {
      seeds[i] = rand();
      if (seeds[i] < 2)
         seeds[i] = 2;
   }

   // allocate DEVICE input / output buffers
   getError(cudaMalloc( (void**) &d_iobuffer, frameBuffer.size() * sizeof(frameBuffer[0])));


   // create DEVICE SEED buffer and propogate on DEVICE
   int sizeBytes = sizeof(unsigned int) * width * height * 2;
   getError(cudaMalloc( (void**) &d_seeds, sizeBytes));
   getError(cudaMemcpy( d_seeds, seeds, sizeBytes, cudaMemcpyHostToDevice));
}

// initCUDA()
// INPUT:   nothing
// RETURN:  nothing
//   perform first allocations for the GPU. CPU function
static void initCUDA() {
   // one time allocate spheres
   int sizeBytes = sizeof(Sphere) * sphereCount;
   getError(cudaMalloc( (void **) &d_spheres, sizeBytes));
   getError(cudaMemcpy(d_spheres, &spheres[0], sizeBytes, cudaMemcpyHostToDevice));

   // one time allocate camera
   sizeBytes = sizeof(Camera);
   getError(cudaMalloc( (void **) &d_camera, sizeBytes));
   getError(cudaMemcpy(d_camera, &camera, sizeBytes, cudaMemcpyHostToDevice));

   // allocate other GPU structures
   allocateBuffers();
}

// executeKernel()
// INPUT:   nothing
// RETURN:  nothing
//   prepares and executes a GPU call. CPU function
static void executeKernel() {
   // prepare kernel call
   dim3 blockDims(512,1,1);
   dim3 gridDims((unsigned int) ceil((double)(width*height/blockDims.x)), 1, 1 );

   // Run the kernel
   if (runTest) colorWheelKernel<<<gridDims, blockDims>>>(d_iobuffer, width, height);
   else radianceKernel<<<gridDims, blockDims>>>(d_iobuffer, d_seeds, d_camera, d_spheres, width, height, sphereCount, currentSample);
}

// updateRendering()
// INPUT:   nothing
// RETURN:  nothing
//   Measure metrics and perform a screen update. CPU function
void updateRendering() {
   // start timer and stats
   double startTime = wallClockTime();
   int startSampleCount = currentSample;

   // copy input from HOST to DEVICE
   getError(cudaMemcpy( d_iobuffer, &frameBuffer[0], frameBuffer.size() * sizeof(frameBuffer[0]), cudaMemcpyHostToDevice ));
   getError(cudaMemcpy( d_seeds, &seeds[0], frameBuffer.size() * sizeof(frameBuffer[0]), cudaMemcpyHostToDevice ));

   // execute CUDA work
   executeKernel();
   currentSample++;

   // copy output from DEVICE to HOST
   getError(cudaMemcpy(&frameBuffer[0], d_iobuffer, frameBuffer.size() * sizeof(frameBuffer[0]), cudaMemcpyDeviceToHost ));
   getError(cudaMemcpy(&seeds[0], d_seeds, sizeof(seeds[0]) * width * height * 2, cudaMemcpyDeviceToHost ));

   // finalize and print stats to screen
   const double elapsedTime = wallClockTime() - startTime;
   const int samples = currentSample - startSampleCount;
   const double sampleSec = samples * height * width / elapsedTime;
   sprintf(captionBuffer, "Rendering time %.3f sec (pass %d)  Sample/sec  %.1fK\n",
      elapsedTime, currentSample, sampleSec / 1000.f);
}

// reInitScene()
// INPUT:   nothing
// RETURN:  nothing
//   Resets sample count and copies scene to GPU memory. CPU function
void reInitScene() {
   // reset sample count
   currentSample = 0;

   // Write scene buffer to DEVICE
   getError(cudaMemcpy(d_spheres, &spheres[0], sizeof(Sphere) * sphereCount, cudaMemcpyHostToDevice));
}

// reInitCamera()
// INPUT:   whether or not to clear and reallocate buffers
// RETURN:  nothing
//   Reinitializes and updates camera position and can clear buffers. CPU function
void reInitCamera(const bool reallocBuffers) {
   // Check if I have to reallocate buffers
   if (reallocBuffers) {
      freeBuffers();
      updateCamera();
      allocateBuffers();
   } else {
      updateCamera();
   }

   // Write camera buffer to DEVICE
   getError(cudaMemcpy(d_camera, &camera, sizeof(Camera), cudaMemcpyHostToDevice));

   // reset samples
   currentSample = 0;
}

// main()
//   Entry point of program. CPU function.
int main(int argc, char *argv[]) {
   // print usage information to user
   fprintf(stderr, "Usage: %s\n", argv[0]);
   fprintf(stderr, "Usage: %s <CUDA test>\n", argv[0]);
   fprintf(stderr, "Usage: %s <window width> <window height> <scene file>\n", argv[0]);

   // read in h, w, and scene file
   if (argc == 4) {
      width = atoi(argv[1]);
      height = atoi(argv[2]);
      readScene(argv[3]);
   }

   // read in if CUDA function test
   // default scene if [0, FALSE] is passed in
   else if (argc == 2) {
      runTest = atoi(argv[1]);

      spheres = CornellSpheres;
      sphereCount = sizeof(CornellSpheres) / sizeof(Sphere);

      vinit(camera.orig, 50.f, 45.f, 205.6f);
      vinit(camera.target, 50.f, 45 - 0.042612f, 204.6);
   }

   // load default scene
   else if (argc == 1) {
      spheres = CornellSpheres;
      sphereCount = sizeof(CornellSpheres) / sizeof(Sphere);

      vinit(camera.orig, 50.f, 45.f, 205.6f);
      vinit(camera.target, 50.f, 45 - 0.042612f, 204.6);
   }

   // otherwise error
   else {
      exit(-1);
   }

   // initialize setings
   updateCamera();
   initCUDA();
   initGlut(argc, argv, title);

   // start render loop
   glutMainLoop();

   return 0;
}
